#!/usr/bin/env python

"""
Generic file access.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.data import *
from array import array
import zlib

# Constants.

CACHE_SIZE = 1000

# Classes.

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.record = array('B')    # record buffer
        self.cache = array('B')
        self.data_start = 0

    def reset(self):

        "To be used to reset the state of the reader or writer between records."

        pass

    def seek(self, offset):
        self.f.seek(offset)
        self.reset()

    def rewind(self):
        self.seek(self.data_start)

    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None

class FileWriter(File):

    "Writing basic data types to files."

    def __init__(self, f):
        File.__init__(self, f)
        self.written = 0

    def tell(self):
        # NOTE: Will not be accurate within the current record.
        return self.written

    def begin_record(self):
        pass

    def end_record(self):
        if self.record:
            length = len(self.record)
            before = len(self.cache)
            vint_to_array(length, self.cache)
            length_size = len(self.cache) - before
            self.cache += self.record
            self.written += length_size + length
            self.record = array('B')
            self.flush()

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        vint_to_array(number, self.record)

    def write_numbers(self, numbers):

        "Write 'numbers' to the file using a variable length encoding."

        for number in numbers:
            vint_to_array(number, self.record)

    def write_string(self, s, compress=0):

        """
        Write 's' to the file, recording its length and compressing the string
        if 'compress' is set to a true value.
        """

        # Convert Unicode objects to strings.

        if isinstance(s, unicode):
            s = s.encode("utf-8")

        # Compress the string if requested.

        if compress:
            cs = zlib.compress(s)

            # Take any shorter than the original.

            if len(cs) < len(s):
                flag = "z"
                s = cs
            else:
                flag = "-"

        else:
            flag = ""

        # Write the length of the data before the data itself.

        length = len(s)
        self.record.fromstring("".join([flag, vint(length), s]))

    def write_sequence_value(self, value, size):
        sequence_to_array(value, size, self.record)

    def write_sequence_values(self, values, size):
        vint_to_array(len(values), self.record)
        for value in values:
            self.write_sequence_value(value, size)

    def write_delta_sequence(self, values, size):
        convert_sequence(values, get_subtractor(values[0]))
        self.write_sequence_values(values, size)

    def write_monotonic_sequence(self, values, size):
        convert_sequence(values, get_monotonic_subtractor(values[0]))
        self.write_sequence_values(values, size)

    def flush(self, force=0):
        if self.f is not None:
            self.end_record()
            if force or len(self.cache) > CACHE_SIZE:
                self.cache.tofile(self.f)
                self.cache = array('B')

    def close(self):
        self.flush(1)
        File.close(self)

class FileReader(File):

    "Reading basic data types from files."

    def __init__(self, f):
        File.__init__(self, f)
        self.record_start = 0
        self.record_end = 0
        self.cache_start = 0
        self.begin()

    def begin(self):

        "Initialise file-wide parameters."

        pass

    def begin_record(self):
        self.start = 0
        try:
            size = self.read_number_from_file()
            self.record = self.from_cache(size)
        except EOFError:
            pass

    def end_record(self):
        pass

    def seek(self, offset):
        if self.cache_start <= offset < self.cache_start + len(self.cache):
            self.cache = self.cache[offset - self.cache_start:]
        else:
            self.f.seek(offset)
            self.cache = array('B')
        self.cache_start = offset
        self.record_start = 0
        self.record_end = 0
        self.reset()

    def tell(self):
        return self.cache_start + self.record_start + self.start

    def ensure_cache(self, size):
        if size > len(self.cache) - self.record_end:
            self.cache = self.cache[self.record_end:]
            self.cache_start += self.record_end
            s = self.f.read(CACHE_SIZE)
            self.cache.fromstring(s)
            self.record_start = 0
            if not s:
                raise EOFError
        else:
            self.record_start = self.record_end
        self.record_end = self.record_start + size

    def from_cache(self, size):
        self.ensure_cache(size)
        return self.cache[self.record_start:self.record_end]

    def read_number_from_file(self):

        "Read a number from the file."

        # Read each byte, adding it to the number.

        a = array('B')
        a += self.from_cache(1)
        csd = a[-1]
        if csd < 128:
            return csd
        else:
            while csd & 128:
                a += self.from_cache(1)
                csd = a[-1]
            return vint_from_array(a)

    def read_number(self):

        "Read a number from the current record."

        n, self.start = vint_from_array_start(self.record, self.start)
        return n

    def read_numbers(self, n):
        l = []
        i = 0
        while i < n:
            l.append(self.read_number())
            i += 1
        return l

    def read_string(self, decompress=0):

        """
        Read a string from the current record, decompressing the stored data if
        'decompress' is set to a true value.
        """

        # Decompress the data if requested.

        if decompress:
            flag = chr(self.record[self.start])
            self.start += 1
        else:
            flag = "-"

        length = self.read_number()
        start = self.start
        self.start += length
        s = self.record[start:self.start].tostring()

        # Perform decompression if applicable.

        if flag == "z":
            s = zlib.decompress(s)

        # Convert strings to Unicode objects.

        return unicode(s, "utf-8")

    def read_sequence_value(self, size):
        value, self.start = sequence_from_array(self.record, size, self.start)
        return value

    def read_sequences(self, size):
        values = []
        length = self.read_number()
        i = 0
        while i < length:
            values.append(self.read_sequence_value(size))
            i += 1
        return values

    def read_delta_sequence(self, size):
        values = self.read_sequences(size)
        convert_sequence(values, get_adder(values[0]))
        return values

    def read_monotonic_sequence(self, size):
        values = self.read_sequences(size)
        convert_sequence(values, get_monotonic_adder(values[0]))
        return values

# vim: tabstop=4 expandtab shiftwidth=4
